/*************************************************************************
This file is part of dres the resource policy dependency resolver.

Copyright (C) 2010 Nokia Corporation.

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation
version 2.1 of the License.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
USA.
*************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include <dres/mm.h>
#include <dres/vm.h>

#include <check.h>


#define STACK_SIZE 50

#define fatal(ec, fmt, args...) do {                                        \
        fprintf(stderr, "%s: fatal error: "fmt"\n", __FUNCTION__, ## args); \
        exit(ec);                                                           \
    } while (0)

#define DEBUG_LEVEL 0
#define debug_printf(args...) do { if (DEBUG_LEVEL > 0)  printf(args); } while(0);

int DBG_VM = 0;


/********************
 * dump_global
 ********************/
void
dump_global(vm_global_t *g)
{
    char *s;
    int   i;
    
    debug_printf("global with %d fact%s\n", g->nfact, g->nfact == 1 ? "" : "s");
    for (i = 0; i < g->nfact; i++) {
        s = ohm_structure_to_string(OHM_STRUCTURE(g->facts[i]));
        debug_printf("  #%d: ", i); if (DEBUG_LEVEL > 0)  vm_fact_print(stdout, g->facts[i]); debug_printf("\n");
    }
}


/********************
 * dump_factstore
 ********************/
void
dump_factstore(void)
{
    char *s;
    
    s = ohm_fact_store_to_string(ohm_fact_store_get_fact_store());
    debug_printf("factstore = %s\n", s);
    g_free(s);
}


/********************
 * stack_pop_all
 ********************/
int
stack_pop_all(vm_stack_t *stack)
{
    int          i, n, type;
    double       d;
    char        *s;
    vm_global_t *g;

    i = 0;
    while ((type = vm_type(stack)) != VM_TYPE_UNKNOWN) {
        switch (type) {
        case VM_TYPE_INTEGER:
            n = vm_pop_int(stack);
            debug_printf("#%d: popped %d (0x%x) from the stack...\n", i, n, n);
            break;
            
        case VM_TYPE_DOUBLE:
            d = vm_pop_double(stack);
            debug_printf("#%d: popped %f from the stack...\n", i, d);
            break;
            
        case VM_TYPE_STRING:
            s = vm_pop_string(stack);
            debug_printf("#%d: popped %s from the stack...\n", i, s);
            break;

        case VM_TYPE_GLOBAL:
            g = vm_pop_global(stack);
            debug_printf("#%d: popped global %p from the stack\n", i, g);
            dump_global(g);
            vm_global_free(g);
            break;
            
        default:
            fatal(EINVAL, "#%d: unexpected object on stack", i);
            return EINVAL;
        }
        
        vm_stack_trim(stack, 0);
        i++;
    }
    
    return 0;
}


/********************
 * stack_test
 ********************/
START_TEST (test_stack)
{
    vm_stack_t *stack  = vm_stack_new(1);
    int         i, err = 0;

    debug_printf("*** [%s] ***\n", __FUNCTION__);

    fail_if (!stack, "failed to allocate a stack");
    
    for (i = 0; i < STACK_SIZE; i++) {
    retry:
        switch (i & 0x3) {
        case 0: err = vm_push_int(stack, i);           break;
        case 1: err = vm_push_string(stack, "foobar"); break;
        case 2: err = vm_push_double(stack, 3.0*i);    break;
        case 3: err = vm_push_string(stack, "barfoo"); break;
        }            
        
        if (err == ENOMEM && !vm_stack_grow(stack, 5))
            goto retry;
        else fail_if (err, "failed to push integer %d on the stack", i);
        debug_printf("push #%d\n", i);
    }

    stack_pop_all(stack);
    vm_stack_del(stack);
}
END_TEST

/********************
 * chunk_test
 ********************/
START_TEST (test_chunk)
{
    vm_stack_t *stack = vm_stack_new(20);
    vm_chunk_t *chunk = vm_chunk_new(200);
    vm_state_t  vm;
    int         err;

    debug_printf("*** [%s] ***\n", __FUNCTION__);

    /* allocate initial stack and code buffer */
    fail_if(stack == NULL || chunk == NULL, "failed to allocate stack and/or code");

    /* generate some code */
    VM_INSTR_PUSH_INT(chunk, cgfail, err, 10);
    VM_INSTR_PUSH_DOUBLE(chunk, cgfail, err, 3.141);
    VM_INSTR_PUSH_INT(chunk, cgfail, err, 0xb19b00b);
    VM_INSTR_PUSH_DOUBLE(chunk, cgfail, err, 9.81);
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "this is a test string...");


    /* execute the code */
    vm.stack = stack;
    vm.chunk = chunk;

    vm.pc     = chunk->instrs;
    vm.ninstr = chunk->ninstr;
    vm.nsize  = chunk->nsize;

    vm_run(&vm);
    stack_pop_all(stack);
    
    vm_stack_del(stack);
    vm_chunk_del(chunk);

    return;

 cgfail:
    fail("code generation failed");
}
END_TEST

/********************
 * filter_test
 ********************/
START_TEST (test_filter)
{
    vm_stack_t *stack = vm_stack_new(20);
    vm_chunk_t *chunk = vm_chunk_new(10);
    vm_state_t  vm;
    int         err;

    debug_printf("*** [%s] ***\n", __FUNCTION__);
    
    fail_if (stack == NULL || chunk == NULL, "failed to allocate stack and/or code");

    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "bar");
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "foo");
    VM_INSTR_PUSH_INT(chunk, cgfail, err, 123);
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "foobar");
    VM_INSTR_CREATE(chunk, cgfail, err, 2);
    
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "bar");
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "foo");
    VM_INSTR_PUSH_INT(chunk, cgfail, err, 123);
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "foobar");
    VM_INSTR_FILTER(chunk, cgfail, err, 2);

    
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "bar");
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "foo");
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "barfoo");
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "foobar");
    VM_INSTR_CREATE(chunk, cgfail, err, 2);
    
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "foo");
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "bar");
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "xyzzy");
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "foobar");
    VM_INSTR_FILTER(chunk, cgfail, err, 2);

    vm.stack = stack;
    vm.chunk = chunk;
    
    vm.pc     = chunk->instrs;
    vm.ninstr = chunk->ninstr;
    vm.nsize  = chunk->nsize;

    vm_run(&vm);

    stack_pop_all(vm.stack);
    
    vm_stack_del(stack);
    vm_chunk_del(chunk);

    return;

 cgfail:
    fail("code generation failed");
}
END_TEST

/********************
 * set_test
 ********************/
START_TEST (test_set)
{
    vm_stack_t *stack = vm_stack_new(20);
    vm_chunk_t *chunk = vm_chunk_new(10);
    vm_state_t  vm;
    int         err;

    debug_printf("*** [%s] ***\n", __FUNCTION__);
    
    fail_if(stack == NULL || chunk == NULL, "failed to allocate stack and/or code");

    /* $a = { foo:'bar', foobar:123 } */
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "bar");
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "foo");
    VM_INSTR_PUSH_INT(chunk, cgfail, err, 123);
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "foobar");
    VM_INSTR_CREATE(chunk, cgfail, err, 2);
    VM_INSTR_PUSH_GLOBAL(chunk, cgfail, err, "a");
    VM_INSTR_SET(chunk, cgfail, err);

    /* $b = { foo: 'bar', foobar: 'barfoo' } */
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "bar");
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "foo");
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "barfoo");
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "foobar");
    VM_INSTR_CREATE(chunk, cgfail, err, 2);
    VM_INSTR_PUSH_GLOBAL(chunk, cgfail, err, "b");
    VM_INSTR_SET(chunk, cgfail, err);

    /* $c = $a */
    VM_INSTR_PUSH_GLOBAL(chunk, cgfail, err, "a");
    VM_INSTR_PUSH_GLOBAL(chunk, cgfail, err, "c");
    VM_INSTR_SET(chunk, cgfail, err);

    /* $d = $b */
    VM_INSTR_PUSH_GLOBAL(chunk, cgfail, err, "b");
    VM_INSTR_PUSH_GLOBAL(chunk, cgfail, err, "d");
    VM_INSTR_SET(chunk, cgfail, err);

    /* $e = $a */
    VM_INSTR_PUSH_GLOBAL(chunk, cgfail, err, "a");
    VM_INSTR_PUSH_GLOBAL(chunk, cgfail, err, "e");
    VM_INSTR_SET(chunk, cgfail, err);
    
    /* $e:foobar = 456 */
    VM_INSTR_PUSH_INT(chunk, cgfail, err, 456);
    VM_INSTR_PUSH_GLOBAL(chunk, cgfail, err, "e");
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "foobar");
    VM_INSTR_SET_FIELD(chunk, cgfail, err);

    vm.stack = stack;
    vm.chunk = chunk;
    
    vm.pc     = chunk->instrs;
    vm.ninstr = chunk->ninstr;
    vm.nsize  = chunk->nsize;

    vm_run(&vm);

    stack_pop_all(vm.stack);

    dump_factstore();
    
    vm_stack_del(stack);
    vm_chunk_del(chunk);

    return;

 cgfail:
    fail("code generation failed");
}
END_TEST


/********************
 * echo_handler
 ********************/
int
echo_handler(void *data, char *name,
             vm_stack_entry_t *args, int narg, vm_stack_entry_t *retval)
{
    int i;

    debug_printf("%s data: %p (%s)\n", __FUNCTION__, data, data ? (char *)data : "");

    for (i = 0; i < narg; i++) {
        switch (args[i].type) {
        case VM_TYPE_INTEGER: debug_printf("%d", args[i].v.i); break;
        case VM_TYPE_DOUBLE:  debug_printf("%f", args[i].v.d); break;
        case VM_TYPE_STRING:  debug_printf("%s", args[i].v.s); break;
        case VM_TYPE_GLOBAL:  if (DEBUG_LEVEL > 0)  vm_global_print(stdout, args[i].v.g); break;
        default:              debug_printf("<unknown type 0x%x>", args[i].type);
        }
        debug_printf(" ");
    }
    debug_printf("\n");
    
    retval->type = VM_TYPE_INTEGER;
    retval->v.i  = 0;
    
    return 0;
    
    (void)name;
}


/********************
 * call_test
 ********************/
START_TEST (test_call)
{
    vm_stack_t *stack = vm_stack_new(20);
    vm_chunk_t *chunk = vm_chunk_new(10);
    vm_state_t  vm;
    int         err;

    debug_printf("*** [%s] ***\n", __FUNCTION__);

    memset(&vm, 0, sizeof(vm));
    
    fail_if(stack == NULL || chunk == NULL, "failed to allocate stack and/or code");

    if ((err = vm_method_add(&vm, "echo", echo_handler, "foo")) != 0)
        fail("failed to register echo handler");

    /* $a = { foo:'bar', foobar:123 } */
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "bar");
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "foo");
    VM_INSTR_PUSH_INT(chunk, cgfail, err, 123);
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "foobar");
    VM_INSTR_CREATE(chunk, cgfail, err, 2);
    VM_INSTR_PUSH_GLOBAL(chunk, cgfail, err, "a");
    VM_INSTR_SET(chunk, cgfail, err);

    /* $b = { foo: 'bar', foobar: 'barfoo' } */
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "bar");
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "foo");
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "barfoo");
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "foobar");
    VM_INSTR_CREATE(chunk, cgfail, err, 2);
    VM_INSTR_PUSH_GLOBAL(chunk, cgfail, err, "b");
    VM_INSTR_SET(chunk, cgfail, err);

    /* $c = $a */
    VM_INSTR_PUSH_GLOBAL(chunk, cgfail, err, "a");
    VM_INSTR_PUSH_GLOBAL(chunk, cgfail, err, "c");
    VM_INSTR_SET(chunk, cgfail, err);

    /* $d = $b */
    VM_INSTR_PUSH_GLOBAL(chunk, cgfail, err, "b");
    VM_INSTR_PUSH_GLOBAL(chunk, cgfail, err, "d");
    VM_INSTR_SET(chunk, cgfail, err);

    /* $e = $a */
    VM_INSTR_PUSH_GLOBAL(chunk, cgfail, err, "a");
    VM_INSTR_PUSH_GLOBAL(chunk, cgfail, err, "e");
    VM_INSTR_SET(chunk, cgfail, err);
    
    /* $e:foobar = 456 */
    VM_INSTR_PUSH_INT(chunk, cgfail, err, 456);
    VM_INSTR_PUSH_GLOBAL(chunk, cgfail, err, "e");
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "foobar");
    VM_INSTR_SET_FIELD(chunk, cgfail, err);

    /* echo($a, $b, $c, $d, $e, 1, 2.0, 'three') */
    VM_INSTR_PUSH_GLOBAL(chunk, cgfail, err, "a");
    VM_INSTR_PUSH_GLOBAL(chunk, cgfail, err, "b");
    VM_INSTR_PUSH_GLOBAL(chunk, cgfail, err, "c");
    VM_INSTR_PUSH_GLOBAL(chunk, cgfail, err, "d");
    VM_INSTR_PUSH_GLOBAL(chunk, cgfail, err, "e");
    VM_INSTR_PUSH_INT(chunk, cgfail, err, 1);
    VM_INSTR_PUSH_DOUBLE(chunk, cgfail, err, 2.0);
    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "three");

    VM_INSTR_PUSH_STRING(chunk, cgfail, err, "echo");
    VM_INSTR_CALL(chunk, cgfail, err, 8);

    vm.stack = stack;
    vm.chunk = chunk;
    
    vm.pc     = chunk->instrs;
    vm.ninstr = chunk->ninstr;
    vm.nsize  = chunk->nsize;

    vm_run(&vm);

    stack_pop_all(vm.stack);

    dump_factstore();
    
    vm_stack_del(stack);
    vm_chunk_del(chunk);

    return;


 cgfail:
    fail("code generation failed");
}
END_TEST


TCase *
vm_case (int test_id)
{
    g_type_init();

    TCase *tc_vm = tcase_create ("vm");
    if (test_id == 0 || test_id == 1)  tcase_add_test (tc_vm, test_stack);
    if (test_id == 0 || test_id == 2)  tcase_add_test (tc_vm, test_chunk);
    //if (test_id == 0 || test_id == 3)  tcase_add_test (tc_vm, test_filter);
    if (test_id == 0 || test_id == 4)  tcase_add_test (tc_vm, test_set);
    //if (test_id == 0 || test_id == 5)  tcase_add_test (tc_vm, test_call);

    return tc_vm;
}





/* 
 * Local Variables:
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 * vim:set expandtab shiftwidth=4:
 */

